package net.pdynet.odbcproxyclient;

import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;

import net.pdynet.odbcproxyclient.jaxb.CloseRequest;
import net.pdynet.odbcproxyclient.jaxb.CloseResponse;
import net.pdynet.odbcproxyclient.jaxb.ConnectRequest;
import net.pdynet.odbcproxyclient.jaxb.ConnectResponse;
import net.pdynet.odbcproxyclient.jaxb.ExecuteRequest;
import net.pdynet.odbcproxyclient.jaxb.ExecuteResponse;
import net.pdynet.odbcproxyclient.jaxb.SelectRequest;
import net.pdynet.odbcproxyclient.jaxb.SelectResponse;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.util.EntityUtils;

public class OdbcProxyClient {
	
	private URI baseUri = null;
	private HttpClient httpClient = null;
	
	public OdbcProxyClient() throws URISyntaxException {
		baseUri = new URI("http://localhost:47115/odbcproxy/");
	}
	
	public ConnectResponse connect(String connectionString, boolean usingOleDb) {
		
		ConnectResponse result = null;
		
		if (StringUtils.isBlank(connectionString))
			throw new IllegalStateException("connectionString is blank.");
		
		try {
			ConnectRequest connectRequest = new ConnectRequest();
			connectRequest.setConnectionString(connectionString);
			connectRequest.setUsingOleDb(usingOleDb);
			
			URI uri = getBaseUri().resolve("connect");
			HttpClient httpClient = getHttpClient();
			HttpPost httpPost = new HttpPost(uri);
			httpPost.setHeader("Accept", "application/xml");
			
			ContentType ct = ContentType.create("application/xml", "UTF-8");
			HttpEntity requestEntity = new JaxbEntity(connectRequest, ct);
			httpPost.setEntity(requestEntity);
			
			HttpResponse response = httpClient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			HttpEntity responseEntity = response.getEntity();
			
			if (statusCode != 200) {
				String responseBody = EntityUtils.toString(responseEntity, "UTF-8");
				throw new OdbcProxyException("Connect operation failed with HTTP status code " + statusCode + "\r\nResponse body:\r\n" + responseBody);
			}
			
			try (InputStream is = responseEntity.getContent()) {
				JAXBContext jaxbContext = JAXBContext.newInstance(ConnectResponse.class);
				Unmarshaller marshaller = jaxbContext.createUnmarshaller();
				result = (ConnectResponse) marshaller.unmarshal(is);
			}
			
			/*
			System.out.println(result.getConnectionID());
			System.out.println(result.getConnectionAutoCloseTime().toGregorianCalendar().getTime());
			System.out.println(result.getError());
			System.out.println(result.isSuccess());
			*/
			
		} catch (OdbcProxyException x) {
			throw x;
		} catch (Exception x) {
			throw new OdbcProxyException("ODBC proxy connect error", x);
		}
		
		return result;
	}
	
	public CloseResponse close(String connectionID) {
		
		CloseResponse result = null;
		
		if (StringUtils.isBlank(connectionID))
			throw new IllegalStateException("connectionID is blank.");
		
		try {
			CloseRequest closeRequest = new CloseRequest();
			closeRequest.setConnectionID(connectionID);
			
			URI uri = getBaseUri().resolve("close");
			HttpClient httpClient = getHttpClient();
			HttpPost httpPost = new HttpPost(uri);
			httpPost.setHeader("Accept", "application/xml");
			
			ContentType ct = ContentType.create("application/xml", "UTF-8");
			HttpEntity requestEntity = new JaxbEntity(closeRequest, ct);
			httpPost.setEntity(requestEntity);
			
			HttpResponse response = httpClient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			HttpEntity responseEntity = response.getEntity();
			
			if (statusCode != 200) {
				String responseBody = EntityUtils.toString(responseEntity, "UTF-8");
				throw new OdbcProxyException("Close operation failed with HTTP status code " + statusCode + "\r\nResponse body:\r\n" + responseBody);
			}
			
			try (InputStream is = responseEntity.getContent()) {
				JAXBContext jaxbContext = JAXBContext.newInstance(CloseResponse.class);
				Unmarshaller marshaller = jaxbContext.createUnmarshaller();
				result = (CloseResponse) marshaller.unmarshal(is);
			}
			
			/*
			System.out.println(result.isSuccess());
			System.out.println(result.getError());
			*/
			
		} catch (OdbcProxyException x) {
			throw x;
		} catch (Exception x) {
			throw new OdbcProxyException("ODBC proxy close error", x);
		} finally {
			if (httpClient != null) {
				httpClient.getConnectionManager().shutdown();
				httpClient = null;
			}
		}
		
		return result;
	}
	
	public SelectResponse select(String connectionID, String query) {
		
		SelectResponse result = null;
		
		if (StringUtils.isBlank(connectionID))
			throw new IllegalStateException("connectionID is blank.");

		if (StringUtils.isBlank(query))
			throw new IllegalStateException("query is blank.");
		
		try {
			SelectRequest selectRequest = new SelectRequest();
			selectRequest.setConnectionID(connectionID);
			selectRequest.setQuery(query);
			
			URI uri = getBaseUri().resolve("select");
			HttpClient httpClient = getHttpClient();
			HttpPost httpPost = new HttpPost(uri);
			httpPost.setHeader("Accept", "application/xml");
			
			ContentType ct = ContentType.create("application/xml", "UTF-8");
			HttpEntity requestEntity = new JaxbEntity(selectRequest, ct);
			httpPost.setEntity(requestEntity);
			
			HttpResponse response = httpClient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			HttpEntity responseEntity = response.getEntity();
			
			if (statusCode != 200) {
				String responseBody = EntityUtils.toString(responseEntity, "UTF-8");
				throw new OdbcProxyException("Select operation failed with HTTP status code " + statusCode + "\r\nResponse body:\r\n" + responseBody);
			}
			
			try (InputStream is = responseEntity.getContent()) {
				JAXBContext jaxbContext = JAXBContext.newInstance(SelectResponse.class);
				Unmarshaller marshaller = jaxbContext.createUnmarshaller();
				result = (SelectResponse) marshaller.unmarshal(is);
			}
			
			/*
			System.out.println(result.isSuccess());
			System.out.println(result.getError());
			System.out.println(result.getConnectionAutoCloseTime());
			*/
			
		} catch (OdbcProxyException x) {
			throw x;
		} catch (Exception x) {
			throw new OdbcProxyException("ODBC proxy select error", x);
		}
		
		return result;
	}
	
	public ExecuteResponse execute(String connectionID, String query) {
		ExecuteResponse result = null;
		
		if (StringUtils.isBlank(connectionID))
			throw new IllegalStateException("connectionID is blank.");

		if (StringUtils.isBlank(query))
			throw new IllegalStateException("query is blank.");
		
		try {
			ExecuteRequest executeRequest = new ExecuteRequest();
			executeRequest.setConnectionID(connectionID);
			executeRequest.setQuery(query);
			
			URI uri = getBaseUri().resolve("execute");
			HttpClient httpClient = getHttpClient();
			HttpPost httpPost = new HttpPost(uri);
			httpPost.setHeader("Accept", "application/xml");
			
			ContentType ct = ContentType.create("application/xml", "UTF-8");
			HttpEntity requestEntity = new JaxbEntity(executeRequest, ct);
			httpPost.setEntity(requestEntity);
			
			HttpResponse response = httpClient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			HttpEntity responseEntity = response.getEntity();
			
			if (statusCode != 200) {
				String responseBody = EntityUtils.toString(responseEntity, "UTF-8");
				throw new OdbcProxyException("Execute operation failed with HTTP status code " + statusCode + "\r\nResponse body:\r\n" + responseBody);
			}
			
			try (InputStream is = responseEntity.getContent()) {
				JAXBContext jaxbContext = JAXBContext.newInstance(ExecuteResponse.class);
				Unmarshaller marshaller = jaxbContext.createUnmarshaller();
				result = (ExecuteResponse) marshaller.unmarshal(is);
			}
			
		} catch (OdbcProxyException x) {
			throw x;
		} catch (Exception x) {
			throw new OdbcProxyException("ODBC proxy execute error", x);
		}
		
		return result;
	}
	
	// -------------------------------------------------
	private HttpClient getHttpClient()
			throws KeyManagementException, UnrecoverableKeyException,
			NoSuchAlgorithmException, KeyStoreException, URISyntaxException {
		
		if (httpClient == null) {
			HttpClientHelper httpClientHelper = new HttpClientHelper();
			httpClient = httpClientHelper.getHttpClient();
		}
		
		return httpClient;
	}
	
	public URI getBaseUri() {
		return baseUri;
	}
	
	public void setBaseUri(URI baseUri) {
		this.baseUri = baseUri;
	}
	
	public void setBaseUri(String baseUri) throws URISyntaxException {
		if (!baseUri.endsWith("/"))
			baseUri += "/";
		
		this.baseUri = new URI(baseUri);
	}
	
}
